@@ -172,19 +172,27 @@
 #define SSL_LOG_ERROR   (1<<1)
 #define SSL_LOG_WARN    (1<<2)
 #define SSL_LOG_INFO    (1<<3)
-#define SSL_LOG_DEBUG   (1<<4)
-#define SSL_LOG_MASK    (SSL_LOG_ERROR|SSL_LOG_WARN|SSL_LOG_INFO|SSL_LOG_DEBUG)
+#define SSL_LOG_TRACE   (1<<4)
+#define SSL_LOG_DEBUG   (1<<5)
+#define SSL_LOG_MASK    (SSL_LOG_ERROR|SSL_LOG_WARN|SSL_LOG_INFO|SSL_LOG_TRACE|SSL_LOG_DEBUG)
 
-#define SSL_ADD_NONE    (1<<8)
-#define SSL_ADD_ERRNO   (1<<9)
-#define SSL_ADD_SSLEAY  (1<<10)
+#define SSL_ADD_NONE     (1<<8)
+#define SSL_ADD_ERRNO    (1<<9)
+#define SSL_ADD_SSLEAY   (1<<10)
+#define SSL_NO_TIMESTAMP (1<<11)
+#define SSL_NO_LEVELID   (1<<12)
+#define SSL_NO_NEWLINE   (1<<13)
 
 /*
  * Defaults for the configuration
  */
 
-#define CFGDEF_CERTIFCATE_FILE        "conf/ssl.crt/server.crt"
-#define CFGDEF_SESSION_CACHE_TIMEOUT  300
+#ifndef SSL_CERTIFCATE_FILE
+#define SSL_CERTIFCATE_FILE        "conf/ssl.crt/server.crt"
+#endif
+#ifndef SSL_SESSION_CACHE_TIMEOUT
+#define SSL_SESSION_CACHE_TIMEOUT  300
+#endif
 
 /* 
  * Set to TRUE if you want to allow SSL to be disabled on a per-request basis
@@ -355,12 +363,11 @@ typedef enum {
 /*
  * Define the Path Checking modes
  */
-typedef enum {
-   SSL_PCM_EXISTS     = 1,
-   SSL_PCM_ISREG      = 2,
-   SSL_PCM_ISDIR      = 4,
-   SSL_PCM_ISNONZERO  = 8
-} PathCheckMode;
+#define SSL_PCM_EXISTS     1
+#define SSL_PCM_ISREG      2
+#define SSL_PCM_ISDIR      4
+#define SSL_PCM_ISNONZERO  8
+typedef unsigned int PathCheckMode;
 
 /*
  * Define the SSL session cache modes and structures
@@ -443,7 +450,6 @@ typedef struct {
     char        *szKeyFile;
     char        *szCACertificatePath;
     char        *szCACertificateFile;
-    char        *szCACertificateReqFile;
     char        *szLogFile;
     char        *szCipherSuite;
     FILE        *fileLogFile;
@@ -492,7 +498,6 @@ const char  *ssl_cmd_SSLCertificateFile(cmd_parms *, char *, char *);
 const char  *ssl_cmd_SSLCertificateKeyFile(cmd_parms *, char *, char *);
 const char  *ssl_cmd_SSLCACertificatePath(cmd_parms *, char *, char *);
 const char  *ssl_cmd_SSLCACertificateFile(cmd_parms *, char *, char *);
-const char  *ssl_cmd_SSLCACertificateReqFile(cmd_parms *, char *, char *);
 const char  *ssl_cmd_SSLVerifyClient(cmd_parms *, char *, char *level);
 const char  *ssl_cmd_SSLVerifyDepth(cmd_parms *, char *, char *);
 const char  *ssl_cmd_SSLSessionCache(cmd_parms *, char *, char *);
@@ -506,8 +511,9 @@ const char  *ssl_cmd_SSLRequire(cmd_parms *, SSLDirConfigRec *, char *);
 /*  module initialization  */
 void         ssl_init_Module(server_rec *, pool *);
 void         ssl_init_SSLeay(server_rec *);
-void         ssl_init_Child(server_rec *, pool *);
 void         ssl_init_GetCertAndKey(server_rec *, pool *, SSLSrvConfigRec *);
+STACK       *ssl_init_FindCAList(server_rec *, pool *, char *, char *);
+void         ssl_init_Child(server_rec *, pool *);
 
 /*  Apache API hooks  */
 void         ssl_hook_AddModule(module *);
@@ -527,6 +533,7 @@ int          ssl_callback_SSLVerify(int, X509_STORE_CTX *);
 int          ssl_callback_NewSessionCacheEntry(SSL *, SSL_SESSION *);
 SSL_SESSION *ssl_callback_GetSessionCacheEntry(SSL *, unsigned char *, int, int *);
 void         ssl_callback_DelSessionCacheEntry(SSL_CTX *, SSL_SESSION *);
+void         ssl_callback_LogTracingState(SSL *, int, int);
 
 /*  Session Cache Support  */
 void         ssl_scache_init(server_rec *, pool *);
@@ -586,6 +593,7 @@ char        *ssl_var_lookup(pool *, server_rec *, conn_rec *, request_rec *, cha
 
 /*  I/O  */
 void         ssl_io_register(void);
+long         ssl_io_data_cb(BIO *, int, char *, int, long, long);
 
 /*  Extensions  */
 void         ssl_ext_register(void);
