@@ -143,7 +143,7 @@ void ssl_init_Module(server_rec *s, pool *p)
         if (sc->nVerifyDepth == UNSET)
             sc->nVerifyDepth = 1;
         if (sc->nSessionCacheTimeout == UNSET)
-            sc->nSessionCacheTimeout = CFGDEF_SESSION_CACHE_TIMEOUT;
+            sc->nSessionCacheTimeout = SSL_SESSION_CACHE_TIMEOUT;
         if (sc->nPassPhraseDialogType == SSL_PPTYPE_UNSET)
             sc->nPassPhraseDialogType = SSL_PPTYPE_BUILTIN;
 
@@ -361,7 +361,7 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
     /*
      *  Create new SSL context and configure callbacks
      */
-    ssl_log(s, SSL_LOG_DEBUG, 
+    ssl_log(s, SSL_LOG_TRACE, 
             "Init: (%s) Creating new SSL context", cpServerName);
     sc->pSSLCtx = SSL_CTX_new(SSLv23_server_method());
     SSL_CTX_set_verify(sc->pSSLCtx, nVerify, 
@@ -374,11 +374,13 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
                  ssl_callback_DelSessionCacheEntry);
     SSL_CTX_set_tmp_rsa_callback(sc->pSSLCtx, 
                  ssl_callback_TmpRSA);
+    SSL_CTX_set_info_callback(sc->pSSLCtx, 
+                 ssl_callback_LogTracingState);
 
     /*
      *  Configure required SSL Ciphers 
      */
-    ssl_log(s, SSL_LOG_DEBUG, 
+    ssl_log(s, SSL_LOG_TRACE, 
             "Init: (%s) Configuring permitted SSL ciphers", cpServerName);
     if (sc->szCipherSuite != NULL) {
         if (!SSL_CTX_set_cipher_list(sc->pSSLCtx, sc->szCipherSuite)) {
@@ -390,47 +392,33 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
     }
 
     /*
-     * Configure SSL Client verification
+     * Configure Client Authentication details
      */
-    ssl_log(s, SSL_LOG_DEBUG, 
-            "Init: (%s) Configuring client verification locations", cpServerName);
-    if (   ( (   sc->szCACertificateFile 
-              || sc->szCACertificatePath)
-           && !SSL_CTX_load_verify_locations(sc->pSSLCtx,
-                                             sc->szCACertificateFile,
-                                             sc->szCACertificatePath) )
-        || !SSL_CTX_set_default_verify_paths(sc->pSSLCtx) ) {
-        ssl_log(s, SSL_LOG_ERROR|SSL_ADD_SSLEAY, 
-                "Init: (%s) Unable to configure SSL verify locations",
-                cpServerName);
-        ssl_die();
+    if (sc->szCACertificateFile != NULL || sc->szCACertificatePath != NULL) {
+        STACK *skCAList;
+        ssl_log(s, SSL_LOG_TRACE, 
+                "Init: (%s) Configuring client authentication", cpServerName);
+        if (!SSL_CTX_load_verify_locations(sc->pSSLCtx, 
+                sc->szCACertificateFile, sc->szCACertificatePath)) {
+            ssl_log(s, SSL_LOG_ERROR|SSL_ADD_SSLEAY, 
+                    "Init: (%s) Unable to configure verify locations "
+                    "for client authentication", cpServerName);
+            ssl_die();
+        }
+        if ((skCAList = ssl_init_FindCAList(s, p, sc->szCACertificateFile, 
+                                            sc->szCACertificatePath)) == NULL) {
+            ssl_log(s, SSL_LOG_ERROR,
+                    "Init: (%s) Unable to determine list of available "
+                    "CA certificates for client authentication", cpServerName);
+            ssl_die();
+        }
+        SSL_CTX_set_client_CA_list(sc->pSSLCtx, skCAList);
     }
 
-    /* 
-     *  Additionally load certificates which will be sent to the
-     *  client on `SSLv3 write certificate request A'. This is
-     *  optionally used by the clients to speedup the server
-     *  authentication when SSLv3 certificate chaining is ised.
-     *  Here the client loads intermediate certificates in the
-     *  chain from the server. This defaults to
-     *  SSLCACertificateFile. 
-     */
-    if (   sc->szCACertificateReqFile != NULL
-        || sc->szCACertificateFile    != NULL)
-        ssl_log(s, SSL_LOG_DEBUG, "Init: (%s) Reading server CA certification file",
-                cpServerName);
-    if (sc->szCACertificateReqFile != NULL)
-        SSL_CTX_set_client_CA_list(sc->pSSLCtx,
-            SSL_load_client_CA_file(sc->szCACertificateReqFile));
-    else
-        if (sc->szCACertificateFile != NULL)
-            SSL_CTX_set_client_CA_list(sc->pSSLCtx,
-                SSL_load_client_CA_file(sc->szCACertificateFile));
-
     /*
      *  Configure server certificate
      */
-    ssl_log(s, SSL_LOG_DEBUG, 
+    ssl_log(s, SSL_LOG_TRACE, 
             "Init: (%s) Configuring server certificate", cpServerName);
     if ((ppX509 = (X509 **)ssl_ds_table_get(mc->tPublicCert, 
                                             cpServerName)) == NULL) {
@@ -443,7 +431,7 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
     /*
      *  Configure server private key 
      */
-    ssl_log(s, SSL_LOG_DEBUG, 
+    ssl_log(s, SSL_LOG_TRACE, 
             "Init: (%s) Configuring server private key", cpServerName);
     if ((ppRSA = (RSA **)ssl_ds_table_get(mc->tPrivateKey, cpServerName)) == NULL) {
         ssl_log(s, SSL_LOG_ERROR, 
@@ -455,6 +443,76 @@ void ssl_init_GetCertAndKey(server_rec *s, pool *p, SSLSrvConfigRec *sc)
     return;
 }
 
+static int ssl_init_FindCAList_X509NameCmp(X509_NAME **a, X509_NAME **b)
+{
+    return(X509_NAME_cmp(*a, *b));
+}
+
+STACK *ssl_init_FindCAList(server_rec *s, pool *pp, char *cpCAfile, char *cpCApath)
+{
+    STACK *skCAList;
+    STACK *sk;
+    DIR *dir;
+    struct DIR_TYPE *direntry;
+    char *cp;
+    pool *p;
+    int n;
+
+    /*
+     * Use a subpool so we don't bloat up the server pool which
+     * is remains in memory for the complete operation time of
+     * the server.
+     */
+    p = ap_make_sub_pool(pp);
+
+    /*
+     * Start with a empty stack/list where new
+     * entries get added in sorted order.
+     */
+    skCAList = sk_new(ssl_init_FindCAList_X509NameCmp);
+
+    /*
+     * Process CA certificate bundle file
+     */
+    if (cpCAfile != NULL) {
+        sk = SSL_load_client_CA_file(cpCAfile);
+        for(n = 0; sk != NULL && n < sk_num(sk); n++) {
+            ssl_log(s, SSL_LOG_TRACE,
+                    "CA certificate: %s",
+                    X509_NAME_oneline((X509_NAME *)sk_value(sk, n), NULL, 0));
+            if (sk_find(skCAList, sk_value(sk, n)) < 0)
+                sk_push(skCAList, sk_value(sk, n));
+        }
+    }
+
+    /*
+     * Process CA certificate path files
+     */
+    if (cpCApath != NULL) {
+        dir = ap_popendir(p, cpCApath);
+        while ((direntry = readdir(dir)) != NULL) {
+            cp = ap_pstrcat(p, cpCApath, "/", direntry->d_name, NULL);
+            sk = SSL_load_client_CA_file(cp);
+            for(n = 0; sk != NULL && n < sk_num(sk); n++) {
+                ssl_log(s, SSL_LOG_TRACE,
+                        "CA certificate: %s",
+                        X509_NAME_oneline((X509_NAME *)sk_value(sk, n), NULL, 0));
+                if (sk_find(skCAList, sk_value(sk, n)) < 0)
+                    sk_push(skCAList, sk_value(sk, n));
+            }
+        }
+        ap_pclosedir(p, dir);
+    }
+
+    /*
+     * Cleanup
+     */
+    sk_set_cmp_func(skCAList, NULL);
+    ap_destroy_pool(p);
+
+    return skCAList;
+}
+
 void ssl_init_Child(server_rec *s, pool *p)
 {
      /* open the mutex lockfile */
