@@ -134,6 +134,7 @@ void ssl_hook_NewConnection(conn_rec *conn)
     SSLSrvConfigRec *sc;
     SSL *ssl;
     char *cp;
+    int rc;
 
     /*
      * Get context
@@ -174,11 +175,20 @@ void ssl_hook_NewConnection(conn_rec *conn)
      * so we can detach later.
      */
     ssl = SSL_new(sc->pSSLCtx);
+    SSL_set_app_data(ssl, srvr);
     SSL_set_fd(ssl, fb->fd);
     ap_ctx_set(fb->ctx, "ssl", ssl);
     ap_register_cleanup(conn->pool, (void *)conn, 
                         ssl_hook_CloseConnection, ssl_hook_CloseConnection);
 
+    /*
+     * Configure SSLeay BIO Data Logging 
+     */
+    if (sc->nLogLevel >= SSL_LOG_DEBUG) {
+        BIO_set_callback(SSL_get_rbio(ssl), ssl_io_data_cb);
+        BIO_set_callback_arg(SSL_get_rbio(ssl), ssl);
+    }
+
     /*
      * Configure the server certificate and private key 
      * which should be used for this connection.
@@ -210,27 +220,45 @@ void ssl_hook_NewConnection(conn_rec *conn)
 
     while (!SSL_is_init_finished(ssl)) {
 
-        if (SSL_accept(ssl) <= 0) {
-            ssl_log(srvr, SSL_LOG_ERROR|SSL_ADD_SSLEAY, 
-                    "SSL handshake failed");
+        if ((rc = SSL_accept(ssl)) <= 0) {
 
-            /*
-             * Check for the case where SSLeay has recognized a HTTP request
-             * line which means the client speaks plain HTTP on our HTTPS
-             * port. Hmmmm...  At least for this error we can be more friendly
-             * and try to provide him with a HTML error page. We have only one
-             * problem: SSLeay has already read some bytes from the HTTP
-             * request. So we have to skip the request line manually and
-             * instead provide a faked one in order to continue the internal
-             * Apache processing.
-             *
-             * (This feature is only available for SSLeay 0.9.0 and higher)
-             */
+            if (SSL_get_error(ssl, rc) == SSL_ERROR_ZERO_RETURN) {
+                /*
+                 * The case where the connection was closed before any data
+                 * was transferred. That's not a real error and can occur
+                 * sporadically with some clients.
+                 */
+                ssl_log(srvr, SSL_LOG_INFO,
+                        "SSL handshake stopped: connection was closed");
+                SSL_set_shutdown(ssl, SSL_RECEIVED_SHUTDOWN); 
+                while (!SSL_shutdown(ssl));
+                SSL_free(ssl);
+                ap_ctx_set(fb->ctx, "ssl", NULL);
+                fb->flags |= B_EOF|B_EOUT;
+                return;
+            }
 #if (SSLEAY_VERSION_NUMBER >= 0x0900)
-            if (ERR_GET_REASON(ERR_peek_error()) == SSL_R_HTTP_REQUEST) {
+            else if (ERR_GET_REASON(ERR_peek_error()) == SSL_R_HTTP_REQUEST) {
+                /*
+                 * The case where SSLeay has recognized a HTTP request:
+                 * This means the client speaks plain HTTP on our HTTPS
+                 * port. Hmmmm...  At least for this error we can be more friendly
+                 * and try to provide him with a HTML error page. We have only one
+                 * problem: SSLeay has already read some bytes from the HTTP
+                 * request. So we have to skip the request line manually and
+                 * instead provide a faked one in order to continue the internal
+                 * Apache processing.
+                 *
+                 * (This feature is only available for SSLeay 0.9.0 and higher)
+                 */
                 char ca[2];
                 int rv;
 
+                /* log the situation */
+                ssl_log(srvr, SSL_LOG_ERROR|SSL_ADD_SSLEAY, 
+                        "SSL handshake failed: HTTP spoken on HTTPS port; "
+                        "trying to send HTML error page");
+
                 /* first: skip the remaining bytes of the request line */
                 do {
                     do {
@@ -240,10 +268,10 @@ void ssl_hook_NewConnection(conn_rec *conn)
 
                 /* second: fake the request line */
                 fb->inbase = ap_palloc(fb->pool, fb->bufsiz);
-                ap_cpystrn(fb->inbase, "GET /mod_ssl:error:HTTP-request HTTP/1.0\r\n", 
+                ap_cpystrn((char *)fb->inbase, "GET /mod_ssl:error:HTTP-request HTTP/1.0\r\n", 
                            fb->bufsiz);
                 fb->inptr = fb->inbase;
-                fb->incnt = strlen(fb->inptr);
+                fb->incnt = strlen((char *)fb->inptr);
 
                 /* third: kick away the SSL stuff */
                 SSL_set_shutdown(ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN); 
@@ -255,20 +283,26 @@ void ssl_hook_NewConnection(conn_rec *conn)
                 return;
             }
 #endif
-
-            /* 
-             * Ok, try to gracefully shutdown the connection:
-             * - send an own shutdown message (be gracefully)
-             * - don't wait for peer's shutdown message (deadloop)
-             * - kick away the SSL stuff immediately
-             * - block the socket, so Apache cannot operate any more
-             */
-            SSL_set_shutdown(ssl, SSL_RECEIVED_SHUTDOWN); 
-            while (!SSL_shutdown(ssl));
-            SSL_free(ssl);
-            ap_ctx_set(fb->ctx, "ssl", NULL);
-            fb->flags |= B_EOF|B_EOUT;
-            return;
+            else {
+                /*
+                 * Ok, anything else is a fatal error
+                 */
+                ssl_log(srvr, SSL_LOG_ERROR|SSL_ADD_SSLEAY, 
+                        "SSL handshake failed");
+                /* 
+                 * try to gracefully shutdown the connection:
+                 * - send an own shutdown message (be gracefully)
+                 * - don't wait for peer's shutdown message (deadloop)
+                 * - kick away the SSL stuff immediately
+                 * - block the socket, so Apache cannot operate any more
+                 */
+                SSL_set_shutdown(ssl, SSL_RECEIVED_SHUTDOWN); 
+                while (!SSL_shutdown(ssl));
+                SSL_free(ssl);
+                ap_ctx_set(fb->ctx, "ssl", NULL);
+                fb->flags |= B_EOF|B_EOUT;
+                return;
+            }
         }
 
         /*
@@ -328,9 +362,14 @@ void ssl_hook_NewConnection(conn_rec *conn)
     }
 
     /* 
-     * This should be a safe optimization...
+     * Improve I/O throughput by using 
+     * SSLeay's read-ahead functionality
+     * (don't used under Win32, because
+     * there we use select())
      */
-    SSL_set_read_ahead(ssl, 1);
+#ifndef WIN32
+    SSL_set_read_ahead(ssl, TRUE);
+#endif
 
     return;
 }
@@ -688,7 +727,7 @@ int ssl_callback_SSLVerify(int ok, X509_STORE_CTX *ctx)
      */
     cp  = X509_NAME_oneline(X509_get_subject_name(xs), NULL, 0);
     cp2 = X509_NAME_oneline(X509_get_issuer_name(xs),  NULL, 0);
-    ssl_log(mc->rCtx.pConn->server, SSL_LOG_DEBUG,
+    ssl_log(mc->rCtx.pConn->server, SSL_LOG_TRACE,
             "Certificate Verification: depth: %d, subject: %s, issuer: %s", 
             errdepth, cp != NULL ? cp : "-unknown-", 
             cp2 != NULL ? cp2 : "-unknown");
@@ -703,7 +742,7 @@ int ssl_callback_SSLVerify(int ok, X509_STORE_CTX *ctx)
             || errnum == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY 
             || errnum == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE  )
         && sc->nVerifyClient == SSL_CVERIFY_OPTIONAL_NO_CA            ) {
-        ssl_log(mc->rCtx.pConn->server, SSL_LOG_DEBUG,
+        ssl_log(mc->rCtx.pConn->server, SSL_LOG_TRACE,
                 "Certificate Verification: Verifiable Issuer is configured as "
                 "optional, therefore we're accepting the certificate");
         ok = TRUE;
@@ -768,7 +807,7 @@ int ssl_callback_NewSessionCacheEntry(SSL *pSSL, SSL_SESSION *pNew)
     /*
      * Log this cache operation
      */
-    ssl_log(mc->rCtx.pConn->server, SSL_LOG_DEBUG, 
+    ssl_log(mc->rCtx.pConn->server, SSL_LOG_TRACE, 
             "Inter-Process Session Cache: request=SET id=%s timeout=%ds (session caching)",
             ssl_scache_id2sz(pNew->session_id, pNew->session_id_length),
             t-time(NULL));
@@ -802,11 +841,11 @@ SSL_SESSION *ssl_callback_GetSessionCacheEntry(
      * Log this cache operation
      */
     if (pSession != NULL)
-        ssl_log(mc->rCtx.pConn->server, SSL_LOG_DEBUG, 
+        ssl_log(mc->rCtx.pConn->server, SSL_LOG_TRACE, 
                 "Inter-Process Session Cache: request=GET status=FOUND id=%s (session reuse)",
                 ssl_scache_id2sz(id, idlen));
     else
-        ssl_log(mc->rCtx.pConn->server, SSL_LOG_DEBUG, 
+        ssl_log(mc->rCtx.pConn->server, SSL_LOG_TRACE, 
                 "Inter-Process Session Cache: request=GET status=MISSED id=%s (session renewal)",
                 ssl_scache_id2sz(id, idlen));
 
@@ -839,10 +878,65 @@ void ssl_callback_DelSessionCacheEntry(
     /*
      * Log this cache operation
      */
-    ssl_log(mc->rCtx.pConn->server, SSL_LOG_DEBUG, 
+    ssl_log(mc->rCtx.pConn->server, SSL_LOG_TRACE, 
             "Inter-Process Session Cache: request=REM status=OK id=%s (session dead)",
             ssl_scache_id2sz(pSession->session_id, pSession->session_id_length));
 
     return;
 }
 
+/*
+ * This callback function is executed while SSLeay processes the
+ * SSL handshake and does SSL record layer stuff. We use it to
+ * trace SSLeay's processing in out SSL logfile.
+ */
+void ssl_callback_LogTracingState(SSL *ssl, int where, int rc)
+{
+    server_rec *s;
+    SSLSrvConfigRec *sc;
+    char *str;
+
+    /* 
+     * find corresponding server
+     */
+    if ((s = (server_rec *)SSL_get_app_data(ssl)) == NULL)
+        return;
+
+    /*
+     * pre-check the logging level to avoid processing overhead
+     */
+    if ((sc = mySrvConfig(s)) == NULL)
+        return;
+    if (sc->nLogLevel < SSL_LOG_TRACE)
+        return;
+
+    /* 
+     * create the various trace messages
+     */
+    if (where & SSL_CB_HANDSHAKE_START)
+        ssl_log(s, SSL_LOG_TRACE, "SSLeay: Handshake: start");
+    else if (where & SSL_CB_HANDSHAKE_DONE)
+        ssl_log(s, SSL_LOG_TRACE, "SSLeay: Handshake: done");
+    else if (where & SSL_CB_LOOP)
+        ssl_log(s, SSL_LOG_TRACE, "SSLeay: Loop: %s", SSL_state_string_long(ssl));
+    else if (where & SSL_CB_READ)
+        ssl_log(s, SSL_LOG_TRACE, "SSLeay: Read: %s", SSL_state_string_long(ssl));
+    else if (where & SSL_CB_WRITE)
+        ssl_log(s, SSL_LOG_TRACE, "SSLeay: Write: %s", SSL_state_string_long(ssl));
+    else if (where & SSL_CB_ALERT) {
+        str = (where & SSL_CB_READ) ? "read" : "write";
+        ssl_log(s, SSL_LOG_TRACE, "SSLeay: Alert: %s:%s\n",
+                SSL_alert_type_string_long(rc), 
+                SSL_alert_desc_string_long(rc));
+    }
+    else if (where & SSL_CB_EXIT) {
+        if (rc == 0)
+            ssl_log(s, SSL_LOG_TRACE, "SSLeay: Exit: failed in %s\n",
+                    SSL_state_string_long(ssl));
+        else if (rc < 0)
+            ssl_log(s, SSL_LOG_TRACE, "SSLeay: Exit: error in %s\n",
+                    SSL_state_string_long(ssl));
+    }
+    return;
+}
+
